package com.stormtales.notevault.data.repositories;

import android.content.Context;
import android.os.Handler;
import android.os.Looper;
import androidx.recyclerview.widget.AsyncListUtil;
import com.stormtales.notevault.data.MusicDatabase;
import com.stormtales.notevault.data.dao.SongDao;
import com.stormtales.notevault.data.entities.NoteSheet;
import com.stormtales.notevault.data.entities.Song;
import com.stormtales.notevault.data.sync.SyncStatus;
import com.stormtales.notevault.network.sync.models.*;
import com.stormtales.notevault.utils.NoteSheetsUtil;

import java.io.File;
import java.io.IOException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;

public class SongSyncRepository {
    private SongDao songDao;
    public SongSyncRepository(Context context) {
        MusicDatabase database = MusicDatabase.getDatabase(context);
        songDao = database.getSongTable();
    }

    public void loadCreatedSongs(LoadDataCallback<List<Song>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<Song> createdSongs = songDao.getSongsBySyncStatus(SyncStatus.CREATED);
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(createdSongs));
        });
    }

    public void loadModifiedSongs(LoadDataCallback<Map<Song, List<NoteSheet>>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            Map<Song, List<NoteSheet>> result = new HashMap<>();
            List<Song> modifiedSongs = songDao.getSongsBySyncStatus(SyncStatus.MODIFIED);
            for(Song song : modifiedSongs) {
                List<NoteSheet> noteSheets = songDao.getNoteSheetsBySong(song.getLocalID());
                result.put(song, noteSheets);
            }
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(result));
        });
    }

    public void loadDeletedSongs(LoadDataCallback<List<Song>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<Song> deletedSongs = songDao.getSongsBySyncStatus(SyncStatus.DELETED);
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(deletedSongs));
        });
    }

    public void markCreatedSongsAsSynced(BatchCreateResponse createdSongs) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<Integer> localIDs = createdSongs.getCreateResponses().stream().map(CreateResponse::getLocalID).collect(Collectors.toList());
            List<Song> requestedSongs = songDao.getSongsByLocalIDs(localIDs);

            for(Song song : requestedSongs) {
                song.setSyncTime(LocalDateTime.now());
                song.setSyncStatus(SyncStatus.SYNCED);

                for(CreateResponse createResponse : createdSongs.getCreateResponses()) {
                    if(createResponse.getLocalID() == song.getLocalID()) {
                        song.setServerID(createResponse.getServerID());
                        break;
                    }
                }
            }
            songDao.updateSongs(requestedSongs);
        });
    }

    public void markSongsAsSynced(List<Song> songs) {
        for(Song song : songs) {
            song.setSyncStatus(SyncStatus.SYNCED);
        }
    }

    public void markModifiedSongsAsSynced(SongModifyBatchResponse modifyBatchResponse) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<String> serverIDs = modifyBatchResponse.getModifiedServerObjects().stream().map(SongModifyResponse::getServerID).collect(Collectors.toList());
            List<Song> requestedSongs = songDao.getSongsByServerIDs(serverIDs);
            for(Song song : requestedSongs) {
                song.setSyncStatus(SyncStatus.SYNCED);
                song.setSyncTime(LocalDateTime.now());


                for(SongModifyResponse modifyResponse : modifyBatchResponse.getModifiedServerObjects()) {
                    if(modifyResponse.getServerID().equals(song.getServerID())) {
                        List<NoteSheet> outdatedNoteSheets = songDao.getNoteSheetsByServerFileNames(modifyResponse.getOutdated_note_sheets());
                        for(NoteSheet noteSheet : outdatedNoteSheets) {
                            noteSheet.setSyncStatus(SyncStatus.MODIFIED);
                        }
                        songDao.updateNoteSheets(outdatedNoteSheets);
                    }
                }
            }


            songDao.updateSongs(requestedSongs);
        });
    }

    public void markDeletedSongsAsSynced(List<String> remoteDeletedSongs, List<Integer> localDeletedSongs) {
        Executors.newSingleThreadExecutor().execute(() -> {
            songDao.deleteSongsByServerIDs(remoteDeletedSongs);
            songDao.deleteSongsByLocalIDs(localDeletedSongs);
        });
    }

    public void getNoteSheetFilesBySongIDs(LoadDataCallback<List<NoteSheet>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<NoteSheet> noteSheets = songDao.getNoteSheetsBySyncStatus(SyncStatus.CREATED);
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(noteSheets));
        });
    }

    public void markCreatedNoteSheetsAsSynced(List<UploadResponse> uploadResponses) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<String> serverSongIDs = uploadResponses.stream().map(UploadResponse::getServerID).collect(Collectors.toList());
            List<Integer> localSongIDs = songDao.getLocalSongIDsByServerIDs(serverSongIDs);

            List<NoteSheet> uploadedNoteSheets = songDao.getNoteSheetsBySongIDs(localSongIDs);
            for(UploadResponse uploadResponse : uploadResponses) {
                for(NoteSheet noteSheet : uploadedNoteSheets) {
                    if(new File(noteSheet.getLocalFileName()).getName().equals(uploadResponse.getLocalFile())) {
                        noteSheet.setServerFileName(uploadResponse.getServerFile());
                        noteSheet.setSyncStatus(SyncStatus.SYNCED);
                        break;
                    }
                }
            }
            songDao.updateNoteSheets(uploadedNoteSheets);
        });
    }

    public void loadModifiedNoteSheets(LoadDataCallback<List<NoteSheet>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<NoteSheet> noteSheets = songDao.getNoteSheetsBySyncStatus(SyncStatus.MODIFIED);
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(noteSheets));
        });
    }

    public void markModifiedNoteSheetsAsSynced(List<UploadResponse> uploadResponses) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<String> serverFileNames = uploadResponses.stream().map(UploadResponse::getServerFile).collect(Collectors.toList());


            List<NoteSheet> noteSheets = songDao.getNoteSheetsByServerFileNames(serverFileNames);
            for(NoteSheet noteSheet : noteSheets) {
                noteSheet.setSyncStatus(SyncStatus.SYNCED);
            }
            songDao.updateNoteSheets(noteSheets);
        });
    }

    public void markSongsAsRemotelyModified(List<String> serverIDs, LoadDataCallback<List<Song>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<Song> songs = songDao.getSongsByServerIDs(serverIDs);
            List<Song> createdSongs = new ArrayList<>();
            for(Song song : songs) {
                song.setSyncStatus(SyncStatus.REMOTE_MODIFIED);
            }

            for(String serverID : serverIDs) {
                boolean foundServerSong = false;
                for(Song song : songs) {
                    if(song.getServerID().equals(serverID)) {
                        foundServerSong = true;
                        break;
                    }
                }

                if(!foundServerSong) {
                    Song song = new Song(serverID);
                    createdSongs.add(song);
                }
            }
            songDao.updateSongs(songs);
            songDao.insertSongs(createdSongs);
            callback.onResult(createdSongs);
        });
    }

    public void loadRemotelyModifiedSongs(LoadDataCallback<List<Song>> callback) {
        Executors.newSingleThreadExecutor().execute(() -> {
            List<Song> songs = songDao.getSongsBySyncStatus(SyncStatus.REMOTE_MODIFIED);
            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(songs));
        });
    }

    public void saveRemoteSongs(List<SongModel> remoteSongs, LoadDataCallback<List<NoteSheet>> callback) {
        Executors.newSingleThreadExecutor().execute(() ->{
            List<String> serverIDs = remoteSongs.stream().map(SongModel::getServerID).collect(Collectors.toList());
            List<Song> songs = songDao.getSongsByServerIDs(serverIDs);
            List<Integer> localSongIDs = songDao.getLocalSongIDsByServerIDs(serverIDs);
            List<NoteSheet> noteSheets = songDao.getNoteSheetsBySongIDs(localSongIDs);

            List<Song> onlyRemoteExistingSongs = new ArrayList<>();
            List<NoteSheet> outdatedNoteSheets = new ArrayList<>();
            List<NoteSheet> onlyRemoteNoteSheets = new ArrayList<>();
            for(SongModel songModel : remoteSongs) {
                boolean found = false;
                for(Song song : songs) {
                    if(song.getServerID().equals(songModel.getServerID())) {
                        found = true;
                        if(!songModel.isDeleted()) {
                           song.setSyncStatus(SyncStatus.SYNCED);
                           song.setSyncTime(LocalDateTime.now());

                           song.setTitle(songModel.getTitle());
                           song.setComposer(songModel.getComposer());
                           song.setGenre(songModel.getGenre());
                           song.setYear(songModel.getYear());

                            for(RemotelyModifiedNoteSheetModel remoteNoteSheet : songModel.getNote_sheets()) {
                                if(noteSheets.isEmpty()) {
                                    NoteSheet noteSheet = new NoteSheet(song.getLocalID(), remoteNoteSheet.getFilename(), remoteNoteSheet.getServer_filename(), null);
                                    onlyRemoteNoteSheets.add(noteSheet);
                                    outdatedNoteSheets.add(noteSheet);
                                } else {
                                    boolean foundNoteSheet = false;
                                    for(NoteSheet noteSheet : outdatedNoteSheets) {
                                        if(remoteNoteSheet.getServer_filename().equals(noteSheet.getServerFileName())) {
                                            foundNoteSheet = true;
                                            if(!remoteNoteSheet.getHash().equals(noteSheet.getHash())) {
                                                outdatedNoteSheets.add(noteSheet);
                                                break;
                                            }
                                        }
                                    }

                                    if(!foundNoteSheet) {
                                        NoteSheet noteSheet = new NoteSheet(song.getLocalID(), remoteNoteSheet.getFilename(), remoteNoteSheet.getServer_filename(), null);
                                        onlyRemoteNoteSheets.add(noteSheet);
                                        outdatedNoteSheets.add(noteSheet);
                                    }
                                }

                            }

                        } else {
                            removeSong(song);
                        }
                    }
                }

                if(!found) {
                    Song song = new Song(songModel);
                    onlyRemoteExistingSongs.add(song);
                }
            }

            songDao.updateSongs(songs);
            songDao.insertSongs(onlyRemoteExistingSongs);
            songDao.insert(onlyRemoteNoteSheets);

            Handler mainHandler = new Handler(Looper.getMainLooper());
            mainHandler.post(()-> callback.onResult(outdatedNoteSheets));
        });
    }

    private void removeSong(Song song) {
        List<NoteSheet> noteSheets = songDao.getNoteSheetsBySong(song.getLocalID());
        for(NoteSheet noteSheet : noteSheets) {
            try {
                NoteSheetsUtil.deleteNoteSheet(noteSheet.getLocalFileName());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
        songDao.deleteNoteSheets(noteSheets);
        songDao.deleteSong(song);

    }



    public void saveUpdatedNoteSheet(String serverFileName, String localFileName, String hash) {
        Executors.newSingleThreadExecutor().execute(()-> {
            NoteSheet noteSheet = songDao.getNoteSheetByServerFileName(serverFileName);
            noteSheet.setHash(hash);
            noteSheet.setLocalFileName(localFileName);
            songDao.updateNoteSheet(noteSheet);
        });
    }

    public interface LoadDataCallback<T> {
        void onResult(T result);
    }
}
